import datetime
import json
from types import NoneType
from typing import Any, Dict, List, Union
import uuid
from basyx.aas import model

import typing
import stringcase

from basyx.aas.model import datatypes
from pydantic import BaseModel, ConfigDict
from pydantic.fields import FieldInfo

from aas_pydantic import aas_model


class AttributeFieldInfo(BaseModel):
    name: str
    field_info: FieldInfo

    model_config = ConfigDict(arbitrary_types_allowed=True)


class AttributeInfo(AttributeFieldInfo):
    value: Any


def get_attribute_field_infos(
    obj: Union[
        type[aas_model.AAS],
        type[aas_model.Submodel],
        type[aas_model.SubmodelElementCollection],
    ]
) -> List[AttributeFieldInfo]:
    """
    Returns a dictionary of all attributes of an object that are not None, do not start with an underscore and are not standard attributes of the aas object.

    Args:
        obj (Union[aas_model.AAS, aas_model.Submodel, aas_model.SubmodelElementCollection]): Object to get the attributes from
    Returns:
        List[AttributeFieldInfo]: List of attributes of the object
    """
    attribute_infos = []
    for attribute_name, field_info in obj.model_fields.items():
        if attribute_name in ["id", "description", "id_short", "semantic_id"]:
            continue
        if attribute_name.startswith("_"):
            continue
        attribute_infos.append(
            AttributeFieldInfo(name=attribute_name, field_info=field_info)
        )
    return attribute_infos


def get_attribute_infos(
    obj: Union[aas_model.AAS, aas_model.Submodel, aas_model.SubmodelElementCollection]
) -> List[AttributeInfo]:
    """
    Returns a dictionary of all attributes of an object that are not None, do not start with an underscore and are not standard attributes of the aas object.

    Args:
        obj (Union[aas_model.AAS, aas_model.Submodel, aas_model.SubmodelElementCollection]): Object to get the attributes from

    Returns:
        List[AttributeInfo]: List of attributes of the object
    """
    attribute_infos = []
    for attribute_name, field_info in obj.model_fields.items():
        if attribute_name in ["id", "description", "id_short", "semantic_id"]:
            continue
        if attribute_name.startswith("_"):
            continue
        attribute_value = getattr(obj, attribute_name)
        attribute_infos.append(
            AttributeInfo(
                name=attribute_name, field_info=field_info, value=attribute_value
            )
        )
    return attribute_infos


def get_str_description(langstring_set: model.LangStringSet) -> str:
    """
    Converts a LangStringSet to a string.
    Args:
        langstring_set (model.LangStringSet): LangStringSet to convert
    Returns:
        str: String representation of the LangStringSet
    """
    if not langstring_set:
        return ""
    if "en" in langstring_set:
        return str(langstring_set.get("en"))
    elif "ger" in langstring_set:
        return str(langstring_set.get("ger"))
    elif "de" in langstring_set:
        return str(langstring_set.get("de"))
    else:
        return str(langstring_set.get(list(langstring_set.keys())[0]))


def get_basyx_description_from_model(
    model_object: (
        aas_model.AAS | aas_model.Submodel | aas_model.SubmodelElementCollection
    ),
) -> model.LangStringSet:
    """
    Creates a LangStringSet from an aas model.
    Args:
        model_object (aas_model.AAS | aas_model.Submodel | aas_model.SubmodelElementCollection): The model to get the description from.
    Returns:
        model.LangStringSet: LangStringSet description representation of the model object
    Raises:
        ValueError: If the description of the model object is not a dict or a string
    """
    if not model_object.description:
        return None
    try:
        dict_description = json.loads(model_object.description)
        if not isinstance(dict_description, dict):
            raise ValueError
    except ValueError:
        dict_description = {"en": model_object.description}
    return model.LangStringSet(dict_description)


def get_class_name_from_basyx_model(
    item: typing.Union[
        model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection
    ]
) -> str:
    """
    Returns the class name of an basyx model from the data specifications.

    Args:
        item (model.HasDataSpecification): Basyx model to get the class name from

    Raises:
        ValueError: If no data specifications are found in the item or if no class name is found

    Returns:
        str: Class name of the basyx model
    """
    if not item.embedded_data_specifications:
        return item.id_short
    for data_spec in item.embedded_data_specifications:
        content = data_spec.data_specification_content
        if not isinstance(content, model.DataSpecificationIEC61360):
            continue
        if not content.preferred_name.get("en") == "class":
            continue
        condition_smc = any(
            key.value == item.id_short for key in data_spec.data_specification.key
        )
        condition_aas_sm = hasattr(item, "id") and any(
            key.value == item.id for key in data_spec.data_specification.key
        )
        if not condition_smc and not condition_aas_sm:
            continue

        return content.value
    raise ValueError(
        f"No class name found in item with id {item.id_short} and type {type(item)}"
    )


def get_class_name_from_basyx_template(
    item: typing.Union[model.Submodel, model.SubmodelElementCollection]
) -> str:
    """
    Returns the class name of an basyx model from the data specifications.

    Args:
        item (model.HasDataSpecification): Basyx model to get the class name from

    Raises:
        ValueError: If no data specifications are found in the item or if no class name is found

    Returns:
        str: Class name of the basyx model
    """
    if not item.embedded_data_specifications:
        return stringcase.camelcase(item.id_short)
    return get_class_name_from_basyx_model(item)


def get_attribute_name_from_basyx_model(
    item: typing.Union[
        model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection
    ],
    referenced_item_id: str,
) -> List[str]:
    """
    Returns the attribute name of the referenced element of the item.

    Args:
        item (typing.Union[model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection]): The container of the refernced item
        referenced_item_id (str): The id of the referenced item

    Raises:
        ValueError: If not data specifications are found in the item or if no attribute name is found

    Returns:
        str: The attribute name of the referenced item
    """
    if not item.embedded_data_specifications:
        return stringcase.snakecase(referenced_item_id)
    attribute_names = []
    for data_spec in item.embedded_data_specifications:
        content = data_spec.data_specification_content
        if not isinstance(content, model.DataSpecificationIEC61360):
            continue
        if not any(
            key.value == referenced_item_id for key in data_spec.data_specification.key
        ):
            continue
        if not content.preferred_name.get("en") == "attribute":
            continue
        attribute_names.append(content.value)
    if attribute_names:
        return attribute_names
    raise ValueError(
        f"Attribute reference to {referenced_item_id} could not be found in {item.id_short} of type {type(item)}"
    )


def get_attribute_names_from_basyx_template(
    item: typing.Union[
        model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection
    ],
    referenced_item_id_short: str,
) -> List[str]:
    """
    Returns the attribute name of the referenced element of the item.

    Args:
        item (typing.Union[model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection]): The container of the refernced item
        referenced_item_id (str): The id of the referenced item

    Raises:
        ValueError: If not data specifications are found in the item or if no attribute name is found

    Returns:
        str: The attribute name of the referenced item
    """
    if not item.embedded_data_specifications:
        return [stringcase.snakecase(referenced_item_id_short)]
    return get_attribute_name_from_basyx_model(item, referenced_item_id_short)


def get_data_specification_for_model_template(
    model_type: typing.Union[
        type[aas_model.AAS],
        type[aas_model.Submodel],
        type[aas_model.SubmodelElementCollection],
    ],
) -> typing.List[model.EmbeddedDataSpecification]:
    return [
        model.EmbeddedDataSpecification(
            data_specification=model.ExternalReference(
                key=(
                    model.Key(
                        type_=model.KeyTypes.GLOBAL_REFERENCE,
                        value=(get_template_id(model_type)),
                    ),
                ),
            ),
            data_specification_content=model.DataSpecificationIEC61360(
                preferred_name=model.LangStringSet({"en": "class"}),
                value=get_template_id(model_type),
            ),
        )
    ]


def get_data_specification_for_model(
    item: typing.Union[
        aas_model.AAS, aas_model.Submodel, aas_model.SubmodelElementCollection
    ],
) -> typing.List[model.EmbeddedDataSpecification]:
    return [
        model.EmbeddedDataSpecification(
            data_specification=model.ExternalReference(
                key=(
                    model.Key(
                        type_=model.KeyTypes.GLOBAL_REFERENCE,
                        value=(
                            item.id
                            if isinstance(
                                item, typing.Union[aas_model.AAS, aas_model.Submodel]
                            )
                            else item.id_short
                        ),
                    ),
                ),
            ),
            data_specification_content=model.DataSpecificationIEC61360(
                preferred_name=model.LangStringSet({"en": "class"}),
                value=item.__class__.__name__.split(".")[-1],
            ),
        )
    ]


def get_model_keys_for_data_specification(
    item: typing.Union[
        NoneType, aas_model.AAS, aas_model.Submodel, aas_model.SubmodelElementCollection
    ] = None,
) -> typing.Tuple[model.Key]:
    if item is None:
        return (
            model.Key(
                type_=model.KeyTypes.GLOBAL_REFERENCE,
                value=uuid.uuid4().hex,
            ),
        )
    return (
        model.Key(
            type_=model.KeyTypes.GLOBAL_REFERENCE,
            value=(
                item.id
                if isinstance(item, typing.Union[aas_model.AAS, aas_model.Submodel])
                else item.id_short
            ),
        ),
    )


def get_data_specification_for_attribute(
    attribute_field_info: AttributeFieldInfo, basyx_attribute: Any
) -> model.EmbeddedDataSpecification:
    model_keys = get_model_keys_for_data_specification(basyx_attribute)
    return model.EmbeddedDataSpecification(
        data_specification=model.ExternalReference(
            key=model_keys,
        ),
        data_specification_content=model.DataSpecificationIEC61360(
            preferred_name=model.LangStringSet({"en": "attribute"}),
            value=attribute_field_info.name,
        ),
    )


def get_optional_data_specification_for_attribute(
    attribute_field_info: AttributeFieldInfo,
) -> typing.Optional[model.EmbeddedDataSpecification]:
    if not (
        typing.get_origin(attribute_field_info.field_info.annotation) is Union
        and type(None) in typing.get_args(attribute_field_info.field_info.annotation)
    ):
        return
    model_keys = get_model_keys_for_data_specification()

    return model.EmbeddedDataSpecification(
        data_specification=model.ExternalReference(
            key=model_keys,
        ),
        data_specification_content=model.DataSpecificationIEC61360(
            preferred_name=model.LangStringSet({"en": "optional"}),
            value=attribute_field_info.name,
        ),
    )


def get_immutable_data_specification_for_attribute_name(
    attribute_name: str,
) -> model.EmbeddedDataSpecification:
    model_keys = get_model_keys_for_data_specification()
    return model.EmbeddedDataSpecification(
        data_specification=model.ExternalReference(
            key=model_keys,
        ),
        data_specification_content=model.DataSpecificationIEC61360(
            preferred_name=model.LangStringSet({"en": "immutable"}),
            value=attribute_name,
        ),
    )


def get_immutable_data_specification_for_attribute(
    attribute_field_info: AttributeFieldInfo,
) -> typing.Optional[model.EmbeddedDataSpecification]:
    if not typing.get_origin(attribute_field_info.field_info.annotation) == tuple:
        return
    return get_immutable_data_specification_for_attribute_name(
        attribute_field_info.name
    )


def get_default_data_specification_for_attribute(
    attribute_field_info: AttributeFieldInfo,
    basyx_attribute: typing.Union[
        NoneType, aas_model.AAS, aas_model.Submodel, aas_model.SubmodelElementCollection
    ],
) -> typing.Optional[model.EmbeddedDataSpecification]:
    model_keys = get_model_keys_for_data_specification(basyx_attribute)
    return model.EmbeddedDataSpecification(
        data_specification=model.ExternalReference(
            key=model_keys,
        ),
        data_specification_content=model.DataSpecificationIEC61360(
            preferred_name=model.LangStringSet({"en": "default"}),
            value=attribute_field_info.field_info.default,
        ),
    )


def get_union_data_specification_for_attribute(
    attribute_field_info: AttributeFieldInfo,
) -> typing.Optional[model.EmbeddedDataSpecification]:
    if not (
        typing.get_origin(attribute_field_info.field_info.annotation) == Union
        and len(
            [
                arg
                for arg in typing.get_args(attribute_field_info.field_info.annotation)
                if arg != NoneType
            ]
        )
        > 1
    ):
        return
    model_keys = get_model_keys_for_data_specification()

    return model.EmbeddedDataSpecification(
        data_specification=model.ExternalReference(
            key=model_keys,
        ),
        data_specification_content=model.DataSpecificationIEC61360(
            preferred_name=model.LangStringSet({"en": "union"}),
            value=attribute_field_info.name,
        ),
    )


def get_default_value_from_basyx_model(
    item: Union[
        model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection
    ],
    attribute_id: str,
) -> typing.Any:
    """
    Returns the default value of an attribute

    Args:
        item (Union[model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection]): The item to check
        attribute_name (str): The name of the attribute

    Returns:
        bool: If the attribute is optional
    """
    if not item.embedded_data_specifications:
        return
    for data_spec in item.embedded_data_specifications:
        content = data_spec.data_specification_content
        if not isinstance(content, model.DataSpecificationIEC61360):
            continue
        if not (content.preferred_name.get("en") == "default"):
            continue
        if not any(
            key.value == attribute_id for key in data_spec.data_specification.key
        ):
            continue
        return content.value
    return


def is_attribute_from_basyx_model_immutable(
    item: typing.Union[
        model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection
    ],
    attribute_name: str,
) -> bool:
    """
    Returns if the referenced item of the item is immutable.

    Args:
        item (typing.Union[model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection]): The container of the refernced item
        referenced_item_id (str): The id of the referenced item

    Raises:
        ValueError: If not data specifications are found in the item or if no attribute name is found

    Returns:
        bool: If the referenced item is immutable
    """
    if not item.embedded_data_specifications:
        return False
    for data_spec in item.embedded_data_specifications:
        content = data_spec.data_specification_content
        if not isinstance(content, model.DataSpecificationIEC61360):
            continue
        if not content.preferred_name.get("en") == "immutable":
            continue
        return content.value == attribute_name
    return False


def is_optional_attribute_type(
    item: Union[
        model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection
    ],
    attribute_name: str,
) -> bool:
    """
    Returns if an attribute of an aas is optional.

    Args:
        item (Union[model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection]): The item to check
        attribute_name (str): The name of the attribute

    Returns:
        bool: If the attribute is optional
    """
    if not item.embedded_data_specifications:
        return True
    for data_spec in item.embedded_data_specifications:
        content = data_spec.data_specification_content
        if not isinstance(content, model.DataSpecificationIEC61360):
            continue
        if not (
            content.preferred_name.get("en") == "optional"
            and content.value == attribute_name
        ):
            continue
        return True
    return False


def is_union_attribute_type(
    item: Union[
        model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection
    ],
    attribute_name: str,
) -> bool:
    """
    Returns if an attribute of an aas is optional.

    Args:
        item (Union[model.AssetAdministrationShell, model.Submodel, model.SubmodelElementCollection]): Aas to get the attribute from
        attribute_name (str): The name of the attribute

    Returns:
        bool: If the attribute is optional
    """
    if not item.embedded_data_specifications:
        return False
    for data_spec in item.embedded_data_specifications:
        content = data_spec.data_specification_content
        if not isinstance(content, model.DataSpecificationIEC61360):
            continue
        if not (
            content.preferred_name.get("en") == "union"
            and content.value == attribute_name
        ):
            continue
        return True
    return False


def get_template_id(
    element: Union[
        type[aas_model.AAS],
        type[aas_model.Submodel],
        type[aas_model.SubmodelElementCollection],
    ]
) -> str:
    return element.__name__.split(".")[-1]


def get_id_short(
    element: Union[
        aas_model.AAS, aas_model.Submodel, aas_model.SubmodelElementCollection
    ]
) -> str:
    if element.id_short:
        return element.id_short
    else:
        return element.id


def get_semantic_id(
    model_object: aas_model.Submodel | aas_model.SubmodelElementCollection,
) -> str | None:
    if model_object.semantic_id:
        semantic_id = model.ExternalReference(
            key=(model.Key(model.KeyTypes.GLOBAL_REFERENCE, model_object.semantic_id),)
        )
    else:
        semantic_id = None
    return semantic_id


def get_value_type_of_attribute(
    attribute: Union[str, int, float, bool]
) -> model.datatypes:
    if isinstance(attribute, bool):
        return model.datatypes.Boolean
    elif isinstance(attribute, int):
        return model.datatypes.Integer
    elif isinstance(attribute, float):
        return model.datatypes.Double
    else:
        return model.datatypes.String


def get_semantic_id_value_of_model(
    basyx_model: typing.Union[model.Submodel, model.SubmodelElement]
) -> str:
    """
    Returns the semantic id of a submodel or submodel element.

    Args:
        basyx_model (model.Submodel | model.SubmodelElement): Basyx model to get the semantic id from.

    Returns:
        str: Semantic id of the model.
    """
    if not isinstance(basyx_model, model.HasSemantics):
        raise NotImplementedError("Type not implemented:", type(basyx_model))
    if not basyx_model.semantic_id:
        return ""
    return basyx_model.semantic_id.key[0].value


def convert_xsdtype_to_primitive_type(
    xsd_data_type: model.DataTypeDefXsd,
) -> aas_model.PrimitiveSubmodelElement:
    if xsd_data_type == datatypes.Duration:
        return str
    elif xsd_data_type == datatypes.DateTime:
        return datetime.datetime
    elif xsd_data_type == datatypes.Date:
        return datetime.datetime
    elif xsd_data_type == datatypes.Time:
        return datetime.time
    # TODO: implement GyearMonth, GYer, GMonthDay, GDay, GMonth
    elif xsd_data_type == datatypes.Boolean:
        return bool
    elif xsd_data_type == datatypes.Base64Binary:
        return bytes
    elif xsd_data_type == datatypes.HexBinary:
        return bytes
    elif xsd_data_type == datatypes.Float:
        return float
    elif xsd_data_type == datatypes.Double:
        return float
    elif xsd_data_type == datatypes.Decimal:
        return float
    elif xsd_data_type == datatypes.Integer:
        return int
    elif xsd_data_type == datatypes.Long:
        return int
    elif xsd_data_type == datatypes.Int:
        return int
    elif xsd_data_type == datatypes.Short:
        return int
    elif xsd_data_type == datatypes.Byte:
        return int
    elif xsd_data_type == datatypes.NonPositiveInteger:
        return int
    elif xsd_data_type == datatypes.NegativeInteger:
        return int
    elif xsd_data_type == datatypes.NonNegativeInteger:
        return int
    elif xsd_data_type == datatypes.PositiveInteger:
        return int
    elif xsd_data_type == datatypes.UnsignedLong:
        return int
    elif xsd_data_type == datatypes.UnsignedInt:
        return int
    elif xsd_data_type == datatypes.UnsignedShort:
        return int
    elif xsd_data_type == datatypes.UnsignedByte:
        return int
    elif xsd_data_type == datatypes.AnyURI:
        return str
    elif xsd_data_type == datatypes.String:
        return str
    elif xsd_data_type == datatypes.NormalizedString:
        return str


def convert_primitive_type_to_xsdtype(
    primitive_type: aas_model.PrimitiveSubmodelElement,
) -> model.DataTypeDefXsd:
    if primitive_type == str:
        return datatypes.String
    elif primitive_type == datetime.datetime:
        return datatypes.DateTime
    elif primitive_type == datetime.time:
        return datatypes.Time
    elif primitive_type == bool:
        return datatypes.Boolean
    elif primitive_type == bytes:
        return datatypes.Base64Binary
    elif primitive_type == float:
        return datatypes.Double
    elif primitive_type == int:
        return datatypes.Integer
    else:
        raise NotImplementedError("Type not implemented:", primitive_type)


def unpatch_id_short_from_temp_attribute(smec: model.SubmodelElementCollection):
    """
    Unpatches the id_short attribute of a SubmodelElementCollection from the temporary attribute.

    Args:
        sm_element (model.SubmodelElementCollection): SubmodelElementCollection to unpatch.
    """
    if not smec.id_short.startswith("generated_submodel_list_hack_"):
        return smec
    no_temp_values = []
    id_short = None
    for sm_element in smec.value:
        if isinstance(sm_element, model.Property) and sm_element.id_short.startswith(
            "temp_id_short_attribute"
        ):
            id_short = sm_element.value
            continue
        no_temp_values.append(sm_element)

    if not id_short:
        # return smec
        new_id_short = smec.parent.id_short
        smec.parent = None
        smec.id_short = new_id_short
        return smec

    for value in no_temp_values:
        smec.value.remove(value)
    new_smec = model.SubmodelElementCollection(
        id_short=id_short,
        value=no_temp_values,
        embedded_data_specifications=smec.embedded_data_specifications,
    )
    return new_smec


def repatch_id_short_to_temp_attribute(
    smec: model.SubmodelElementCollection, temp_smec: model.SubmodelElementCollection
):
    """
    Repatches the id_short attribute of a SubmodelElementCollection to the temporary attribute.

    Args:
        sm_element (model.SubmodelElementCollection): SubmodelElementCollection to repatch.
    """
    values_to_repatch = []
    for sm_element in temp_smec.value:
        values_to_repatch.append(sm_element)
    temp_smec.value.clear()
    for sm_element in values_to_repatch:
        smec.value.add(sm_element)
    return smec



from __future__ import annotations

from types import NoneType
from typing import Annotated, Any, List, Optional, TypeVar, Union
import typing

from basyx.aas.model import AssetAdministrationShell, DictObjectStore, Submodel
from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    ValidationError,
    field_validator,
    model_validator,
)


BasyxModels = AssetAdministrationShell | Submodel | DictObjectStore


def string_does_start_with_a_character(v: str):
    assert v, "value must not be an empty string"
    assert v[0].isalpha(), "value must start with a character"
    return v


AasIdString = Annotated[str, BeforeValidator(string_does_start_with_a_character)]
Reference = TypeVar(
    "Reference",
    bound=Annotated[str, BeforeValidator(string_does_start_with_a_character)],
)


class Referable(BaseModel):
    """
    Base class for all referable classes of the AAS meta model. A Referable is an object with a local id (id_short) and a description.

    Args:
        id_short (IdString): Local id of the object.
        description (str): Description of the object. Defaults to None.
    """

    id_short: AasIdString
    description: str = ""


class Identifiable(Referable):
    """
    Base class for all identifiable classes in the AAS Meta model. An Identifiable is a Referable with a global id (id_).

    Args:
        id (str): Global id of the object.
        id_short (str): Local id of the object.
        description (str, optional): Description of the object. Defaults to None.
    """

    id: AasIdString

    @model_validator(mode="before")
    @classmethod
    def check_id_and_id_short(cls, data: Any) -> Any:
        if isinstance(data, BaseModel):
            data = data.model_dump()
        elif not isinstance(data, dict):
            data = {
                "id": getattr(data, "id", ""),
                "id_short": getattr(data, "id_short", ""),
            }
        assert "id" in data or "id_short" in data, "Either id or id_short must be set"
        if "id_short" not in data:
            data["id_short"] = data["id"]
        if "id" not in data:
            data["id"] = data["id_short"]
        return data


class HasSemantics(BaseModel):
    """
    Base class for all classes that have semantics of the AAS meta model. Semantics are defined by a semantic id, which reference the semantic definition of the object.

    Args:
        semantic_id (str): Semantic id of the object. Defaults to None.
    """

    semantic_id: str = ""


class AAS(Identifiable):
    """
    Base class for all Asset Administration Shells (AAS).

    Args:
        id (str): Global id of the object.
        id_short (str): Local id of the object.
        description (str, optional): Description of the object. Defaults to None.
    """

    @model_validator(mode="before")
    @classmethod
    def set_optional_fields_to_None(cls, data):
        if isinstance(data, BaseModel):
            data = data.model_dump()
        for field_name, field_info in cls.model_fields.items():
            if field_name in data:
                continue
            if typing.get_origin(field_info.annotation) == Union and type(
                None
            ) in typing.get_args(field_info.annotation):
                data[field_name] = None
        return data

    @model_validator(mode="after")
    def check_submodels(self) -> Any:
        for field_name, field_info in self.model_fields.items():
            if field_name in ["id", "id_short", "description"]:
                continue
            elif (
                typing.get_origin(field_info.annotation) == Union
                and type(None) in typing.get_args(field_info.annotation)
                and getattr(self, field_name) is None
            ):
                continue
            try:
                Submodel.model_validate(getattr(self, field_name))
            except ValidationError:
                assert (
                    False
                ), f"All attributes of an AAS must be of type Submodel or inherit from Submodel"
        return self


def is_valid_submodel_element(submodel_element: Any) -> bool:
    if isinstance(submodel_element, NoneType):
        return True
    if isinstance(submodel_element, PrimitiveSubmodelElement):
        return True
    elif isinstance(submodel_element, SubmodelElementCollection):
        return True
    elif (
        isinstance(submodel_element, list)
        or isinstance(submodel_element, tuple)
        or isinstance(submodel_element, set)
    ):
        return all(is_valid_submodel_element(element) for element in submodel_element)
    elif isinstance(submodel_element, Operation):
        return True
    elif isinstance(submodel_element, File):
        return True
    elif isinstance(submodel_element, Blob):
        return True
    try:
        SubmodelElementCollection.model_validate(submodel_element)
        return True
    except:
        return False


class SubmodelElementCollection(HasSemantics, Referable):
    """
    Base class for all submodel element collections.

    Args:
        id_short (str): Local id of the object.
        description (str, optional): Description of the object. Defaults to None.
        semantic_id (str, optional): Semantic id of the object. Defaults to None.
    """

    @model_validator(mode="after")
    def check_submodel_elements(self) -> Any:
        for field_name in self.model_fields:
            if field_name in ["id", "id_short", "description", "semantic_id"]:
                continue
            assert is_valid_submodel_element(
                getattr(self, field_name)
            ), f"All attributes of a SubmodelElementCollection must be valid SubmodelElements. Field {field_name} is not valid."
        return self


class Operation(HasSemantics, Referable):
    input_variables: List[SubmodelElement]
    output_variables: List[SubmodelElement]
    inoutput_variables: List[SubmodelElement]


class File(HasSemantics, Referable):
    media_type: str
    path: str


class Blob(HasSemantics, Referable):
    media_type: str
    content: Optional[bytes] = None


PrimitiveSubmodelElement = int | float | str | bool | bytes
SubmodelElement = (
    PrimitiveSubmodelElement
    | SubmodelElementCollection
    | List["SubmodelElement"]
    | Operation
    | Blob
    | File
)


class Submodel(HasSemantics, Identifiable):
    """
    Base class for all submodels.

    Args:
        id (str): Global id of the object.
        id_short (str): Local id of the object.
        description (str, optional): Description of the object. Defaults to None.
        semantic_id (str, optional): Semantic id of the object. Defaults to None.
    """

    @model_validator(mode="after")
    def check_submodel_elements(self) -> Any:
        for field_name in self.model_fields:
            if field_name in ["id", "id_short", "description", "semantic_id"]:
                continue
            assert is_valid_submodel_element(
                getattr(self, field_name)
            ), f"All attributes of a Submodel must be valid SubmodelElements."
        return self



from aas_pydantic.aas_model import AAS, Submodel, SubmodelElementCollection, Reference
from aas_pydantic.convert_aas_instance import (
    convert_object_store_to_pydantic_models,
    convert_aas_to_pydantic_model_instance,
    convert_submodel_to_model_instance,
)
from aas_pydantic.convert_aas_template import (
    convert_object_store_to_pydantic_types,
    convert_aas_to_pydantic_type,
    convert_submodel_template_to_pydatic_type,
)
from aas_pydantic.convert_pydantic_model import (
    convert_model_to_aas,
    convert_model_to_submodel,
)
from aas_pydantic.convert_pydantic_type import (
    convert_model_to_aas_template,
    convert_model_to_submodel_template,
    convert_model_instance_to_submodel_template,
)



from __future__ import annotations

from collections import OrderedDict
from urllib import parse
from enum import Enum
import uuid


from basyx.aas import model

from typing import Optional, Union
from aas_pydantic import convert_util, aas_model

from aas_pydantic.convert_util import (
    convert_primitive_type_to_xsdtype,
    get_attribute_infos,
    get_id_short,
    get_semantic_id,
    get_value_type_of_attribute,
)

import logging

logger = logging.getLogger(__name__)


def convert_model_to_aas(
    model_aas: aas_model.AAS,
) -> model.DictObjectStore[model.Identifiable]:
    """
    Convert a model aas to an Basyx AssetAdministrationShell and return it as a DictObjectStore with all Submodels

    Args:
        model_aas (aas_model.AAS): model aas to convert

    Returns:
        model.DictObjectStore[model.Identifiable]: DictObjectStore with all Submodels
    """
    aas_attribute_infos = get_attribute_infos(model_aas)
    aas_submodels = {}
    aas_submodel_data_specifications = []
    for attribute_info in aas_attribute_infos:
        submodel = convert_model_to_submodel(model_submodel=attribute_info.value)
        attribute_data_specification = (
            convert_util.get_data_specification_for_attribute(attribute_info, submodel)
        )
        aas_submodel_data_specifications.append(attribute_data_specification)
        if submodel and not submodel.id_short in aas_submodels:
            aas_submodels.update({submodel.id_short: submodel})

    asset_information = model.AssetInformation(
        global_asset_id=model.Identifier(model_aas.id),
        asset_kind=model.AssetKind.INSTANCE,
        asset_type=model.Identifier("Instance"),
    )

    basyx_aas = model.AssetAdministrationShell(
        asset_information=asset_information,
        id_short=get_id_short(model_aas),
        id_=model.Identifier(model_aas.id),
        description=convert_util.get_basyx_description_from_model(model_aas),
        submodel={
            model.ModelReference.from_referable(submodel)
            for submodel in aas_submodels.values()
        },
        embedded_data_specifications=convert_util.get_data_specification_for_model(
            model_aas
        )
        + aas_submodel_data_specifications,
    )
    obj_store: model.DictObjectStore[model.Identifiable] = model.DictObjectStore()
    obj_store.add(basyx_aas)
    for sm in aas_submodels.values():
        obj_store.add(sm)
    return obj_store


def convert_model_to_submodel(
    model_submodel: aas_model.Submodel,
) -> Optional[model.Submodel]:
    if not model_submodel:
        return
    submodel_attributes = get_attribute_infos(model_submodel)
    submodel_elements = []
    submodel_element_data_specifications = []

    for attribute_info in submodel_attributes:
        submodel_element = create_submodel_element(
            attribute_info.name, attribute_info.value
        )
        attribute_data_specification = (
            convert_util.get_data_specification_for_attribute(
                attribute_info, submodel_element
            )
        )
        submodel_element_data_specifications.append(attribute_data_specification)
        immutable_attribute_data_specification = (
            convert_util.get_immutable_data_specification_for_attribute(attribute_info)
        )
        if immutable_attribute_data_specification:
            submodel_element_data_specifications.append(
                immutable_attribute_data_specification
            )
        if not attribute_info.field_info.is_required():
            default_data_specification = (
                convert_util.get_default_data_specification_for_attribute(
                    attribute_info, submodel_element
                )
            )
            submodel_element_data_specifications.append(default_data_specification)
        if submodel_element:
            submodel_elements.append(submodel_element)

    basyx_submodel = model.Submodel(
        id_short=get_id_short(model_submodel),
        id_=model.Identifier(model_submodel.id),
        description=convert_util.get_basyx_description_from_model(model_submodel),
        embedded_data_specifications=convert_util.get_data_specification_for_model(
            model_submodel
        )
        + submodel_element_data_specifications,
        semantic_id=get_semantic_id(model_submodel),
        submodel_element=submodel_elements,
    )
    return basyx_submodel


def create_submodel_element(
    attribute_name: str,
    attribute_value: Union[
        aas_model.SubmodelElementCollection, str, float, int, bool, tuple, list, set
    ],
) -> Optional[model.SubmodelElement]:
    """
    Create a basyx SubmodelElement from a model SubmodelElementCollection or a primitive type

    Args:
        attribute_name (str): Name of the attribute that is used for ID and id_short
        attribute_value (Union[ aas_model.SubmodelElementCollection, str, float, int, bool, tuple, list, set ]): Value of the attribute


    Returns:
        model.SubmodelElement: basyx SubmodelElement
    """
    if not attribute_value:
        return
    if isinstance(attribute_value, aas_model.SubmodelElementCollection):
        smc = create_submodel_element_collection(attribute_value)
        return smc
    elif isinstance(attribute_value, (list, tuple, set)):
        sml = create_submodel_element_list(attribute_name, attribute_value)
        return sml
    elif (isinstance(attribute_value, str)) and (
        (
            parse.urlparse(attribute_value).scheme
            and parse.urlparse(attribute_value).netloc
        )
        or (attribute_value.split("_")[-1] in ["id", "ids"])
    ):
        key = model.Key(
            type_=model.KeyTypes.ASSET_ADMINISTRATION_SHELL,
            value=attribute_value,
        )
        reference = model.ModelReference(key=(key,), type_="")
        reference_element = model.ReferenceElement(
            id_short=attribute_name,
            value=reference,
        )
        return reference_element
    elif isinstance(attribute_value, aas_model.File):
        return create_file(attribute_value)
    elif isinstance(attribute_value, aas_model.Blob):
        return create_blob(attribute_value)
    else:
        property = create_property(attribute_name, attribute_value)

        return property


def create_property(
    attribute_name: str,
    attribute_value: Union[str, int, float, bool],
) -> model.Property:
    if isinstance(attribute_value, Enum):
        attribute_value = attribute_value.value
        attribute_type = str
    else:
        attribute_type = type(attribute_value)

    property = model.Property(
        id_short=attribute_name,
        value_type=convert_primitive_type_to_xsdtype(attribute_type),
        value=attribute_value,
    )
    return property


def create_submodel_element_collection(
    model_sec: aas_model.SubmodelElementCollection,
) -> model.SubmodelElementCollection:
    value = []
    smc_attributes = get_attribute_infos(model_sec)
    submodel_element_data_specifications = []

    for attribute_info in smc_attributes:
        sme = create_submodel_element(attribute_info.name, attribute_info.value)
        attribute_data_specification = (
            convert_util.get_data_specification_for_attribute(attribute_info, sme)
        )
        submodel_element_data_specifications.append(attribute_data_specification)
        immutable_attribute_data_specification = (
            convert_util.get_immutable_data_specification_for_attribute(attribute_info)
        )
        if immutable_attribute_data_specification:
            submodel_element_data_specifications.append(
                immutable_attribute_data_specification
            )
        if (
            not attribute_info.field_info.is_required()
            and attribute_info.field_info.default
        ):
            default_data_specification = (
                convert_util.get_default_data_specification_for_attribute(
                    attribute_info, sme
                )
            )
            submodel_element_data_specifications.append(default_data_specification)
        if sme:
            value.append(sme)

    id_short = get_id_short(model_sec)

    smc = model.SubmodelElementCollection(
        id_short=id_short,
        value=value,
        description=convert_util.get_basyx_description_from_model(model_sec),
        embedded_data_specifications=convert_util.get_data_specification_for_model(
            model_sec
        )
        + submodel_element_data_specifications,
        semantic_id=get_semantic_id(model_sec),
    )
    return smc


def patch_id_short_with_temp_attribute(
    submodel_element_collection: model.SubmodelElementCollection,
) -> None:
    """
    Patch the id_short of a SubmodelElementCollection as an attribute in the value of the SubmodelElementCollection, to make it accesible after retrieving from the value list.

    Args:
        submodel_element_collection (model.SubmodelElementCollection): SubmodelElementCollection to patch
    """
    temp_id_short_property = model.Property(
        id_short="temp_id_short_attribute_" + uuid.uuid4().hex,
        value_type=get_value_type_of_attribute(str),
        value=submodel_element_collection.id_short,
    )
    submodel_element_collection.value.add(temp_id_short_property)


def create_submodel_element_list(
    attribute_name: str, value: list | tuple | set
) -> model.SubmodelElementList:
    submodel_elements = []
    submodel_element_ids = OrderedDict()
    for el in value:
        submodel_element = create_submodel_element(attribute_name, el)
        if isinstance(submodel_element, model.SubmodelElementCollection):
            if submodel_element.id_short in submodel_element_ids:
                raise ValueError(
                    f"Submodel element collection with id {submodel_element.id_short} already exists in list"
                )
            submodel_element_ids.update({submodel_element.id_short: None})
            patch_id_short_with_temp_attribute(submodel_element)
        submodel_element.id_short = None
        submodel_elements.append(submodel_element)

    if submodel_elements and isinstance(submodel_elements[0], model.Property):
        value_type_list_element = type(value.__iter__().__next__())
        type_value_list_element = type(submodel_elements[0])
    elif submodel_elements and isinstance(
        submodel_elements[0], model.Reference | model.SubmodelElementCollection
    ):
        value_type_list_element = None
        type_value_list_element = type(submodel_elements[0])
    else:
        value_type_list_element = convert_primitive_type_to_xsdtype(str)
        type_value_list_element = model.Property
    if isinstance(value, set):
        ordered = False
        iterable_type = "set"
    elif isinstance(value, tuple):
        ordered = True
        iterable_type = "tuple"
    elif isinstance(value, list):
        ordered = True
        iterable_type = "list"
    else:
        raise ValueError(
            f"Value must be a list, tuple or set, provided type {type(value)}"
        )

    sml = model.SubmodelElementList(
        id_short=f"{iterable_type}_{uuid.uuid4().hex}",
        type_value_list_element=type_value_list_element,
        value_type_list_element=value_type_list_element,
        value=submodel_elements,
        order_relevant=ordered,
    )
    return sml


def create_file(attribute_value: aas_model.File) -> model.File:
    """
    Function generates a basyx file objects from a pydantic File.

    Args:
        attribute_value (aas_model.File): pydantic File instance.

    Returns:
        model.File: Basyx file.
    """
    return model.File(
        id_short=attribute_value.id_short,
        description=attribute_value.description,
        semantic_id=attribute_value.semantic_id,
        content_type=attribute_value.media_type,
        value=attribute_value.path,
    )


def create_blob(attribute_value: aas_model.Blob) -> model.Blob:
    """
    Function generates a basyx file objects from a pydantic File.

    Args:
        attribute_value (aas_model.File): pydantic File instance.

    Returns:
        model.File: Basyx file.
    """
    return model.Blob(
        id_short=attribute_value.id_short,
        description=attribute_value.description,
        semantic_id=attribute_value.semantic_id,
        content_type=attribute_value.media_type,
        value=attribute_value.content,
    )



from typing import Any, Dict, Optional
import copy


def normalize_schema(
    schema: Dict[str, Any], reference_schemas: Dict[str, Any]
) -> Dict[str, Any]:
    """Recursively normalize schema for comparison, sorting keys and normalizing references."""
    if isinstance(schema, dict):
        if "examples" in schema:
            del schema["examples"]
        if "title" in schema:
            del schema["title"]

        # TODO: remove later when tuple incorrection of length of entries is working when writing to basyx AAS...
        if "minItems" in schema:
            del schema["minItems"]
        if "maxItems" in schema:
            del schema["maxItems"]
        if "prefixItems" in schema:
            schema["items"] = schema["prefixItems"][0]
            del schema["prefixItems"]

        # TODO: remove later when fixed number of string values (enums, literals) is working when writing to basyx AAS based on concept descriptions...
        if "$ref" in schema:
            references_name = schema["$ref"].split("/")[-1]
            if (
                references_name in reference_schemas
                and "enum" in reference_schemas[references_name]
            ):
                schema["title"] = reference_schemas[references_name]["title"]
                schema["type"] = reference_schemas[references_name]["type"]
                del schema["$ref"]
        if "enum" in schema:
            del schema["enum"]

        # Sort the dictionary keys and normalize nested objects
        try:
            return {
                key: normalize_schema(value, reference_schemas=reference_schemas)
                for key, value in sorted(schema.items())
            }
        except TypeError:
            return {
                key: normalize_schema(value, reference_schemas=reference_schemas)
                for key, value in schema.items()
            }
    elif isinstance(schema, list):
        # Sort lists for consistent ordering
        try:
            return sorted(
                normalize_schema(item, reference_schemas=reference_schemas)
                for item in schema
            )
        except TypeError:
            # sort based on values not on key
            schema_list = [
                normalize_schema(item, reference_schemas=reference_schemas)
                for item in schema
            ]
            return sorted(schema_list, key=lambda x: str(x))
    return schema


def compare_properties(
    schema1: Dict[str, Any],
    schema2: Dict[str, Any],
    reference_schemas: Dict[str, Any],
    ignore_tuple_type_hints: bool = False,
) -> bool:
    """Compare properties and required attributes of two schemas."""
    # Check if both schemas define properties
    properties1 = schema1.get("properties", {})
    properties2 = schema2.get("properties", {})

    # Normalize and compare properties
    normalized_schema1 = normalize_schema(
        properties1, reference_schemas=reference_schemas
    )
    normalized_schema2 = normalize_schema(
        properties2, reference_schemas=reference_schemas
    )
    keys_to_update_in_properties2 = {}
    for key, value in normalized_schema1.items():
        if not key in normalized_schema2:
            print("missing key", key)
        if value != normalized_schema2[key]:
            print("different values", key)
            print(value)
            print(normalized_schema2[key])
            keys_to_update_in_properties2[key] = value["items"]
    if ignore_tuple_type_hints:
        for key, value in keys_to_update_in_properties2.items():
            print(
                "updating key",
                key,
                "items from",
                properties2[key]["items"],
                "to",
                value,
            )
            properties2[key]["items"] = value
    if normalize_schema(
        properties1, reference_schemas=reference_schemas
    ) != normalize_schema(properties2, reference_schemas=reference_schemas):
        return False

    # Compare required fields
    required1 = sorted(schema1.get("required", []))
    required2 = sorted(schema2.get("required", []))

    if required1 != required2:
        return False

    return True


def compare_references(
    schema1: Dict[str, Any], schema2: Dict[str, Any], reference_schemas: Dict[str, Any]
) -> bool:
    """Compare references between two schemas recursively, checking definitions."""
    ref1 = schema1.get("$ref")
    ref2 = schema2.get("$ref")

    # If both schemas reference something, ensure they reference the same thing
    if ref1 != ref2:
        return False

    # If references exist, compare the referenced schemas recursively
    if ref1 and ref2:
        ref_schema1 = reference_schemas.get(ref1)
        ref_schema2 = reference_schemas.get(ref2)
        if ref_schema1 and ref_schema2:
            return compare_schemas(ref_schema1, ref_schema2, reference_schemas)

    return True


def compare_schemas(
    schema1: Dict[str, Any],
    schema2: Dict[str, Any],
    reference_schemas: Optional[Dict[str, Any]] = None,
    ignore_tuple_type_hints: bool = False,
) -> bool:
    """Compare two JSON schemas recursively, including properties, references, and required fields."""
    if not reference_schemas:
        reference_schemas = {}
        reference_schemas.update(copy.deepcopy(schema1.get("$defs", {})))
        reference_schemas.update(copy.deepcopy(schema2.get("$defs", {})))

    normalized_schema1 = normalize_schema(schema1, reference_schemas=reference_schemas)
    normalized_schema2 = normalize_schema(schema2, reference_schemas=reference_schemas)

    if normalized_schema1.get("type") != normalized_schema2.get("type"):
        return False

    if not compare_properties(
        normalized_schema1,
        normalized_schema2,
        reference_schemas=reference_schemas,
        ignore_tuple_type_hints=ignore_tuple_type_hints,
    ):
        return False

    if not compare_references(
        normalized_schema1, normalized_schema2, reference_schemas
    ):
        return False

    return True



from __future__ import annotations

from collections import OrderedDict
import json
from types import NoneType
import typing
from urllib import parse
from enum import Enum
import uuid


from basyx.aas import model

from typing import Any, Optional, Set, Union
from pydantic import BaseModel, ConfigDict

# from aas_middleware.model.core import Reference
from aas_pydantic import convert_util, aas_model

from aas_pydantic.convert_util import (
    AttributeFieldInfo,
    convert_primitive_type_to_xsdtype,
    get_attribute_field_infos,
    get_template_id,
)

import basyx.aas.adapter.json.json_serialization
import logging

logger = logging.getLogger(__name__)


def convert_model_to_aas_template(
    model_type: type[aas_model.AAS],
) -> model.DictObjectStore[model.Identifiable]:
    """
    Convert a model aas to an Basyx AssetAdministrationShell and return it as a DictObjectStore with all Submodels

    Args:
        model_type (type[aas_model.AAS]): Type of the model

    Returns:
        model.DictObjectStore[model.Identifiable]: DictObjectStore with all Submodels
    """
    aas_attribute_infos = get_attribute_field_infos(model_type)
    aas_submodels = {}
    aas_submodel_data_specifications = []
    for attribute_info in aas_attribute_infos:
        if typing.get_origin(attribute_info.field_info.annotation) == Union:
            types_to_check = [
                type_annotation
                for type_annotation in typing.get_args(
                    attribute_info.field_info.annotation
                )
                if type_annotation != NoneType
            ]
            optional_attribute_data_specification = (
                convert_util.get_optional_data_specification_for_attribute(
                    attribute_info
                )
            )
            if optional_attribute_data_specification:
                aas_submodel_data_specifications.append(
                    optional_attribute_data_specification
                )
            union_attribute_data_specification = (
                convert_util.get_union_data_specification_for_attribute(attribute_info)
            )
            if union_attribute_data_specification:
                aas_submodel_data_specifications.append(
                    union_attribute_data_specification
                )

        else:
            types_to_check = [attribute_info.field_info.annotation]

        for type_annotation in types_to_check:
            submodel = convert_model_to_submodel_template(model_type=type_annotation)
            attribute_data_specifications = (
                convert_util.get_data_specification_for_attribute(
                    attribute_info, submodel
                )
            )
            aas_submodel_data_specifications.append(attribute_data_specifications)
            if not attribute_info.field_info.is_required():
                default_data_specification = (
                    convert_util.get_default_data_specification_for_attribute(
                        attribute_info, submodel
                    )
                )
                aas_submodel_data_specifications.append(default_data_specification)

            if submodel and not submodel.id_short in aas_submodels:
                aas_submodels.update({submodel.id_short: submodel})

    asset_information = model.AssetInformation(
        asset_kind=model.AssetKind.TYPE,
        asset_type=model.Identifier("Type"),
        global_asset_id=model.Identifier(get_template_id(model_type)),
    )

    basyx_aas = model.AssetAdministrationShell(
        asset_information=asset_information,
        id_short=get_template_id(model_type),
        id_=model.Identifier(get_template_id(model_type)),
        description={
            "en": f"Type aas with id {get_template_id(model_type)} that contains submodel templates"
        },
        submodel={
            model.ModelReference.from_referable(submodel)
            for submodel in aas_submodels.values()
        },
        embedded_data_specifications=convert_util.get_data_specification_for_model_template(
            model_type
        )
        + aas_submodel_data_specifications,
    )
    obj_store: model.DictObjectStore[model.Identifiable] = model.DictObjectStore()
    obj_store.add(basyx_aas)
    for sm in aas_submodels.values():
        obj_store.add(sm)
    return obj_store


def convert_model_instance_to_submodel_template(
    model_instance: aas_model.Submodel,
) -> Optional[model.Submodel]:
    return convert_model_to_submodel_template(type(model_instance))


def convert_model_to_submodel_template(
    model_type: type[aas_model.Submodel],
) -> Optional[model.Submodel]:
    if not model_type:
        return
    submodel_attributes = get_attribute_field_infos(model_type)
    submodel_elements = []
    submodel_element_data_specifications = []

    for attribute_info in submodel_attributes:
        if typing.get_origin(attribute_info.field_info.annotation) == Union:
            types_to_check = [
                type_annotation
                for type_annotation in typing.get_args(
                    attribute_info.field_info.annotation
                )
                if type_annotation != NoneType
            ]
            optional_attribute_data_specification = (
                convert_util.get_optional_data_specification_for_attribute(
                    attribute_info
                )
            )
            if optional_attribute_data_specification:
                submodel_element_data_specifications.append(
                    optional_attribute_data_specification
                )
            union_attribute_data_specification = (
                convert_util.get_union_data_specification_for_attribute(attribute_info)
            )
            if union_attribute_data_specification:
                submodel_element_data_specifications.append(
                    union_attribute_data_specification
                )

        else:
            types_to_check = [attribute_info.field_info.annotation]

        for counter, type_annotation in enumerate(types_to_check):
            if len(types_to_check) > 1:
                attribute_name = f"{attribute_info.name}_{counter}"
            else:
                attribute_name = attribute_info.name
            submodel_element = create_submodel_element_template(
                attribute_name=attribute_name, attribute_type=type_annotation
            )
            attribute_data_specifications = (
                convert_util.get_data_specification_for_attribute(
                    attribute_info, submodel_element
                )
            )
            submodel_element_data_specifications.append(attribute_data_specifications)
            immutable_attribute_data_specification = (
                convert_util.get_immutable_data_specification_for_attribute(
                    attribute_info
                )
            )
            if immutable_attribute_data_specification:
                submodel_element_data_specifications.append(
                    immutable_attribute_data_specification
                )
            if not attribute_info.field_info.is_required():
                default_data_specification = (
                    convert_util.get_default_data_specification_for_attribute(
                        attribute_info, submodel_element
                    )
                )
                submodel_element_data_specifications.append(default_data_specification)
            if submodel_element and not any(
                stored_submodel_element.id_short == submodel_element.id_short
                for stored_submodel_element in submodel_elements
            ):
                submodel_elements.append(submodel_element)

    basyx_submodel = model.Submodel(
        id_short=get_template_id(model_type),
        id_=model.Identifier(get_template_id(model_type)),
        # description=convert_util.get_basyx_description_from_model(model_type)=convert_util.get_basyx_description_from_model(model_type),
        description={
            "en": f"Submodel with id {get_template_id(model_type)} that contains submodel elements"
        },
        embedded_data_specifications=convert_util.get_data_specification_for_model_template(
            model_type
        )
        + submodel_element_data_specifications,
        semantic_id="",
        submodel_element=submodel_elements,
    )
    return basyx_submodel


def create_submodel_element_template(
    attribute_name: str,
    attribute_type: Union[
        type[aas_model.SubmodelElementCollection],
        type[str],
        type[float],
        type[int],
        type[bool],
        type[tuple],
        type[list],
        type[set],
    ],
) -> Optional[model.SubmodelElement]:
    """
    Create a basyx SubmodelElement from a model SubmodelElementCollection or a primitive type

    Args:
        attribute_name (str): Name of the attribute that is used for ID and id_short
        attribute_type (Union[type[aas_model.SubmodelElementCollection], type[str], type[float], type[int], type[bool], type[tuple], type[list], type[set]): Type of the attribute


    Returns:
        model.SubmodelElement: basyx SubmodelElement
    """
    if not attribute_type:
        return
    if (
        typing.get_origin(attribute_type) == list
        or typing.get_origin(attribute_type) == tuple
        or typing.get_origin(attribute_type) == set
    ):
        sml = create_submodel_element_list(attribute_name, attribute_type)
        return sml
    elif attribute_type == aas_model.Reference:
        key = model.Key(
            type_=model.KeyTypes.ASSET_ADMINISTRATION_SHELL,
            value=get_template_id(attribute_type),
        )
        reference = model.ModelReference(key=(key,), type_="")
        reference_element = model.ReferenceElement(
            id_short=attribute_name,
            value=reference,
        )
        return reference_element
    elif typing.get_origin(attribute_type) is typing.Literal:
        property = create_property(attribute_name, str)
        return property
    elif issubclass(attribute_type, Enum):
        property = create_property(attribute_name, str)
        return property
    elif issubclass(attribute_type, aas_model.SubmodelElementCollection):
        smc = create_submodel_element_collection(attribute_type)
        return smc
    elif issubclass(attribute_type, aas_model.File):
        return create_file(attribute_type)
    elif issubclass(attribute_type, aas_model.Blob):
        return create_blob(attribute_type)
    else:
        property = create_property(attribute_name, attribute_type)

        return property


def create_property(
    attribute_name: str,
    attribute_type: Union[type[str], type[float], type[int], type[bool]],
) -> model.Property:
    if issubclass(attribute_type, Enum):
        attribute_type = str

    property = model.Property(
        id_short=attribute_name,
        value_type=convert_primitive_type_to_xsdtype(attribute_type),
        value=None,
    )
    return property


def create_submodel_element_collection(
    model_sec: type[aas_model.SubmodelElementCollection],
) -> model.SubmodelElementCollection:
    value = []
    smc_attributes = get_attribute_field_infos(model_sec)
    submodel_element_data_specifications = []

    for attribute_info in smc_attributes:
        if typing.get_origin(attribute_info.field_info.annotation) == Union:
            types_to_check = [
                type_annotation
                for type_annotation in typing.get_args(
                    attribute_info.field_info.annotation
                )
                if type_annotation != NoneType
            ]
            optional_attribute_data_specification = (
                convert_util.get_optional_data_specification_for_attribute(
                    attribute_info
                )
            )
            if optional_attribute_data_specification:
                submodel_element_data_specifications.append(
                    optional_attribute_data_specification
                )
            union_attribute_data_specification = (
                convert_util.get_union_data_specification_for_attribute(attribute_info)
            )
            if union_attribute_data_specification:
                submodel_element_data_specifications.append(
                    union_attribute_data_specification
                )
        elif attribute_info.field_info.annotation == NoneType:
            continue
        else:
            types_to_check = [attribute_info.field_info.annotation]

        for counter, type_annotation in enumerate(types_to_check):
            if len(types_to_check) > 1:
                attribute_name = f"{attribute_info.name}_{counter}"
            else:
                attribute_name = attribute_info.name
            submodel_element = create_submodel_element_template(
                attribute_name=attribute_name, attribute_type=type_annotation
            )
            attribute_data_specifications = (
                convert_util.get_data_specification_for_attribute(
                    attribute_info, submodel_element
                )
            )
            submodel_element_data_specifications.append(attribute_data_specifications)
            immutable_attribute_data_specification = (
                convert_util.get_immutable_data_specification_for_attribute(
                    attribute_info
                )
            )
            if immutable_attribute_data_specification:
                submodel_element_data_specifications.append(
                    immutable_attribute_data_specification
                )
            if not attribute_info.field_info.is_required():
                default_data_specification = (
                    convert_util.get_default_data_specification_for_attribute(
                        attribute_info, submodel_element
                    )
                )
                submodel_element_data_specifications.append(default_data_specification)
            if submodel_element and not any(
                stored_submodel_element.id_short == submodel_element.id_short
                for stored_submodel_element in value
            ):
                value.append(submodel_element)

    id_short = get_template_id(model_sec)

    smc = model.SubmodelElementCollection(
        id_short=id_short,
        value=value,
        # description=convert_util.get_basyx_description_from_model(model_sec),
        description={
            "en": f"Submodel element collection with id {id_short} that contains submodel elements"
        },
        embedded_data_specifications=convert_util.get_data_specification_for_model_template(
            model_sec
        )
        + submodel_element_data_specifications,
        semantic_id="",
    )
    return smc


def patch_id_short_with_temp_attribute(
    submodel_element_collection: model.SubmodelElementCollection,
) -> None:
    """
    Patch the id_short of a SubmodelElementCollection as an attribute in the value of the SubmodelElementCollection, to make it accesible after retrieving from the value list.

    Args:
        submodel_element_collection (model.SubmodelElementCollection): SubmodelElementCollection to patch
    """
    temp_id_short_property = model.Property(
        id_short="temp_id_short_attribute_" + uuid.uuid4().hex,
        value_type=convert_primitive_type_to_xsdtype(str),
        value=submodel_element_collection.id_short,
    )
    submodel_element_collection.value.add(temp_id_short_property)


def create_submodel_element_list(
    name: str, attribute_type: Union[type[tuple], type[list], type[set]]
) -> model.SubmodelElementList:
    submodel_elements = []
    submodel_element_ids = OrderedDict()
    for el in typing.get_args(attribute_type):
        # TODO: potentially check here because of Unions and Optional types inside lists and sets...
        submodel_element = create_submodel_element_template(name, el)
        if isinstance(submodel_element, model.SubmodelElementCollection):
            if submodel_element.id_short in submodel_element_ids:
                raise ValueError(
                    f"Submodel element collection with id {submodel_element.id_short} already exists in list"
                )
            submodel_element_ids.update({submodel_element.id_short: None})
            patch_id_short_with_temp_attribute(submodel_element)
        submodel_element.id_short = None
        submodel_elements.append(submodel_element)

    if submodel_elements and isinstance(submodel_elements[0], model.Property):
        if len(typing.get_args(attribute_type)) > 1 and not all(
            arg is typing.get_args(attribute_type)[0]
            for arg in typing.get_args(attribute_type)
        ):
            raise ValueError(
                f"Submodel element list with different types is not supported. Please use a SubmodelElementCollection instead."
            )
        value_type_list_element = submodel_elements[0].value_type
        type_value_list_element = type(submodel_elements[0])
    elif submodel_elements and isinstance(
        submodel_elements[0],
        model.Reference
        | model.SubmodelElementCollection
        | model.ReferenceElement
        | model.SubmodelElementList,
    ):
        value_type_list_element = None
        type_value_list_element = type(submodel_elements[0])
    else:
        value_type_list_element = convert_primitive_type_to_xsdtype(str)
        type_value_list_element = model.Property

    if typing.get_origin(attribute_type) == set:
        ordered = False
        iterable_type = "set"
    elif typing.get_origin(attribute_type) == tuple:
        ordered = True
        iterable_type = "tuple"
    elif typing.get_origin(attribute_type) == list:
        ordered = True
        iterable_type = "list"
    else:
        raise ValueError(
            f"Type {attribute_type} is not supported for SubmodelElementList, provided subclass of list, tuple or set"
        )

    sml = model.SubmodelElementList(
        id_short=f"{iterable_type}_of_{get_template_id(typing.get_args(attribute_type)[0])}",
        type_value_list_element=type_value_list_element,
        value_type_list_element=value_type_list_element,
        value=submodel_elements,
        order_relevant=ordered,
    )
    return sml


def create_file(attribute_type: type[aas_model.File]) -> model.File:
    """
    Function generates a basyx file objects from a pydantic File.

    Args:
        attribute_value (aas_model.File): pydantic File instance.

    Returns:
        model.File: Basyx file.
    """
    return model.File(id_short=get_template_id(attribute_type), content_type="unknown")


def create_blob(attribute_type: type[aas_model.Blob]) -> model.Blob:
    """
    Function generates a basyx file objects from a pydantic File.

    Args:
        attribute_value (aas_model.File): pydantic File instance.

    Returns:
        model.File: Basyx file.
    """
    return model.Blob(id_short=get_template_id(attribute_type), content_type="unknown")



from __future__ import annotations

import logging
import typing
from pydantic import BaseModel, Field, create_model
from pydantic_core import PydanticUndefined

# from aas_middleware.model.core import Reference
from aas_pydantic import aas_model
from basyx.aas import model


from aas_pydantic import convert_util
from aas_pydantic.convert_util import (
    convert_xsdtype_to_primitive_type,
    get_semantic_id_value_of_model,
    repatch_id_short_to_temp_attribute,
    unpatch_id_short_from_temp_attribute,
)


def convert_object_store_to_pydantic_types(
    obj_store: model.DictObjectStore,
) -> typing.List[type[aas_model.AAS]]:
    """
    Converts an object store with AAS and submodels to pydantic models, representing the original data structure.

    Args:
        obj_store (model.DictObjectStore): Object store with AAS and submodels

    Returns:
        typing.List[aas_model.AAS]: List of pydantic models
    """
    pydantic_submodel_types: typing.List[aas_model.Submodel] = []
    submodels_pydantic_type_mapping = {}
    for identifiable in obj_store:
        if isinstance(identifiable, model.Submodel):
            pydantic_submodel = convert_submodel_template_to_pydatic_type(identifiable)
            pydantic_submodel_types.append(pydantic_submodel)
            submodels_pydantic_type_mapping[identifiable.id] = (
                pydantic_submodel,
                identifiable,
            )

    pydantic_aas_list: typing.List[type[aas_model.AAS]] = []
    for identifiable in obj_store:
        if isinstance(identifiable, model.AssetAdministrationShell):
            pydantic_aas = convert_aas_to_pydantic_type(
                identifiable, submodels_pydantic_type_mapping
            )
            pydantic_aas_list.append(pydantic_aas)

    return pydantic_aas_list


def convert_aas_to_pydantic_type(
    aas: model.AssetAdministrationShell,
    pydantic_submodel_types: typing.Dict[
        str, typing.Tuple[type[aas_model.Submodel], model.Submodel]
    ],
) -> type[aas_model.AAS]:
    """
    Converts an AAS to a Pydantic model.

    Args:
        aas (model.AssetAdministrationShell): AAS to convert

    Returns:
        aas_model.AAS: Pydantic model of the asset administration shell
    """
    aas_class_name = convert_util.get_class_name_from_basyx_template(aas)
    dict_dynamic_model_creation = get_initial_dict_for_dynamic_model_creation(aas)
    aas_submodel_ids = [sm.get_identifier() for sm in aas.submodel]

    for submodel_id in aas_submodel_ids:
        pydantic_submodel_type, basyx_submodel = pydantic_submodel_types[submodel_id]
        attribute_names_of_submodel = (
            convert_util.get_attribute_names_from_basyx_template(
                aas, basyx_submodel.id_short
            )
        )
        for attribute_name in attribute_names_of_submodel:
            attribute_type = pydantic_submodel_type
            optional = convert_util.is_optional_attribute_type(aas, attribute_name)
            union = convert_util.is_union_attribute_type(aas, attribute_name)
            if optional:
                attribute_type = typing.Optional[attribute_type]
            if attribute_type is None:
                logging.warning(
                    f"Could not convert submodel {submodel_id} to Pydantic model. Skipping."
                )
                continue
            if union and attribute_name in dict_dynamic_model_creation:
                previous_attribute_type = typing.get_args(
                    dict_dynamic_model_creation[attribute_name]
                )[0]
                attribute_type = typing.Union[previous_attribute_type, attribute_type]
            dict_dynamic_model_creation.update(
                {attribute_name: typing.Annotated[attribute_type, Field(examples=[])]}
            )
    model_type = create_model(
        aas_class_name, **dict_dynamic_model_creation, __base__=aas_model.AAS
    )
    return model_type


def get_submodel_element_type(
    sm_element: model.SubmodelElement, immutable: bool = False
) -> type[aas_model.SubmodelElement]:
    """
    Returns the value of a SubmodelElement.

    Args:
        sm_element (model.SubmodelElement): SubmodelElement to get the value from.

    Returns:
        aas_model.SubmodelElement: Value of the SubmodelElement.
    """
    if isinstance(sm_element, model.SubmodelElementCollection):
        return convert_submodel_collection_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.SubmodelElementList):
        return convert_submodel_list_to_pydantic_model(sm_element, immutable)
    elif isinstance(sm_element, model.ReferenceElement):
        return convert_reference_element_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.RelationshipElement):
        return convert_relationship_element_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.Property):
        return convert_property_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.MultiLanguageProperty):
        return convert_multi_language_property_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.File):
        return convert_file_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.Blob):
        return convert_blob_to_pydantic_model(sm_element)
    else:
        raise NotImplementedError("Type not implemented:", type(sm_element))


def get_dynamic_model_creation_dict_from_submodel_element(
    attribute_name: str,
    attribute_type: typing.Type[typing.Any],
    default_value: typing.Any = None,
) -> typing.Dict[str, typing.Any]:
    """
    Converts a SubmodelElement to a dict.

    Args:
        attribute_name (str): Name of the attribute to create in the dictionary.
        sm_element (model.SubmodelElement): SubmodelElement to convert.

    Returns:
        dict: Dictionary that can be used to create a Pydantic model, with Annoated types for the attributes and examples.
    """
    if not default_value:
        default_value = PydanticUndefined
    return {
        attribute_name: typing.Annotated[
            attribute_type, Field(examples=[], default=default_value)
        ]
    }


def get_model_instantiation_dict_from_submodel_element(
    attribute_name: str, attribute_value: typing.Any
) -> typing.Dict[str, typing.Any]:
    """
    Converts a SubmodelElement to a dict.

    Args:
        attribute_name (str): Name of the attribute to create in the dictionary.
        sm_element (model.SubmodelElement): SubmodelElement to convert.

    Returns:
        dict: Dictionary that can be used to instantiate a Pydantic model.
    """
    if isinstance(attribute_value, BaseModel):
        attribute_value = attribute_value.model_dump()
    elif isinstance(attribute_value, (list, set, tuple)) and any(
        isinstance(element, BaseModel) for element in attribute_value
    ):
        attribute_value = [element.model_dump() for element in attribute_value]
    return {attribute_name: attribute_value}


def get_initial_dict_for_dynamic_model_creation(
    basyx_model: (
        model.Submodel
        | model.AssetAdministrationShell
        | model.SubmodelElementCollection
    ),
) -> typing.Dict[str, typing.Any]:
    """
    Returns a dictionary that can be used to create a Pydantic model based on a provided basyx submodel.

    Args:
        basyx_model (model.Submodel | model.AssetAdministrationShell | model.SubmodelElementCollection): Basyx model to create the dictionary from.

    Returns:
        typing.Dict[str, typing.Any]: Dictionary that can be used to create a Pydantic model.
    """
    # model_creation_dict = {
    #     "id_short": typing.Annotated[str, Field(examples=[basyx_model.id_short])],
    #     "description": typing.Annotated[
    #         str,
    #         Field(examples=[convert_util.get_str_description(basyx_model.description)]),
    #     ],
    # }
    # if isinstance(basyx_model, model.Identifiable):
    #     model_creation_dict["id"] = typing.Annotated[
    #         str, Field(examples=[str(basyx_model.id)])
    #     ]
    # if isinstance(basyx_model, model.HasSemantics):
    #     model_creation_dict["semantic_id"] = typing.Annotated[
    #         str, Field(examples=[get_semantic_id_value_of_model(basyx_model)])
    #     ]
    # return model_creation_dict
    return {}


def convert_submodel_template_to_pydatic_type(
    sm: model.Submodel,
) -> type[aas_model.Submodel]:
    """
    Converts a Submodel to a Pydantic model.

    Args:
        sm (model.Submodel): Submodel to convert.

    Returns:
        aas_model.Submodel: Pydantic model of the submodel.
    """
    class_name = convert_util.get_class_name_from_basyx_template(sm)
    dict_dynamic_model_creation = get_initial_dict_for_dynamic_model_creation(sm)

    for sm_element in sm.submodel_element:
        attribute_names = convert_util.get_attribute_names_from_basyx_template(
            sm, sm_element.id_short
        )
        for attribute_name in attribute_names:
            optional = convert_util.is_optional_attribute_type(sm, attribute_name)
            union = convert_util.is_union_attribute_type(sm, attribute_name)
            immutable = convert_util.is_attribute_from_basyx_model_immutable(
                sm, attribute_name
            )
            default_value = convert_util.get_default_value_from_basyx_model(
                sm, sm_element.id_short
            )
            attribute_type = get_submodel_element_type(sm_element, immutable)

            if optional:
                attribute_type = typing.Optional[attribute_type]
            if attribute_type is None:
                logging.warning(
                    f"Could not convert submodel element {attribute_name} to Pydantic model. Skipping."
                )
                continue
            if union and attribute_name in dict_dynamic_model_creation:
                attribute_type = typing.Union[
                    dict_dynamic_model_creation[attribute_name], attribute_type
                ]
            sme_model_creation_dict = (
                get_dynamic_model_creation_dict_from_submodel_element(
                    attribute_name, attribute_type, default_value
                )
            )
            dict_dynamic_model_creation.update(sme_model_creation_dict)
    model_type = create_model(
        class_name, **dict_dynamic_model_creation, __base__=aas_model.Submodel
    )
    return model_type


def convert_submodel_collection_to_pydantic_model(
    sm_element: model.SubmodelElementCollection,
) -> type[aas_model.SubmodelElementCollection]:
    """
    Converts a SubmodelElementCollection to a Pydantic model.

    Args:
        sm_element (model.SubmodelElementCollection): SubmodelElementCollection to convert.

    Returns:
        aas_model.SubmodelElementCollection: Pydantic model of the submodel element collection.
    """
    class_name = convert_util.get_class_name_from_basyx_template(sm_element)
    dict_dynamic_model_creation = get_initial_dict_for_dynamic_model_creation(
        sm_element
    )

    for sub_sm_element in sm_element.value:
        attribute_names = convert_util.get_attribute_names_from_basyx_template(
            sm_element, sub_sm_element.id_short
        )
        for attribute_name in attribute_names:
            optional = convert_util.is_optional_attribute_type(
                sm_element, attribute_name
            )
            union = convert_util.is_union_attribute_type(sm_element, attribute_name)
            immutable = convert_util.is_attribute_from_basyx_model_immutable(
                sm_element, attribute_name
            )
            default_value = convert_util.get_default_value_from_basyx_model(
                sm_element, sub_sm_element.id_short
            )
            attribute_type = get_submodel_element_type(sub_sm_element, immutable)

            if optional:
                attribute_type = typing.Optional[attribute_type]
            if attribute_type is None:
                logging.warning(
                    f"Could not convert submodel element {attribute_name} to Pydantic model. Skipping."
                )
                continue
            if union and attribute_name in dict_dynamic_model_creation:
                attribute_type = typing.Union[
                    dict_dynamic_model_creation[attribute_name], attribute_type
                ]
            dict_sme = get_dynamic_model_creation_dict_from_submodel_element(
                attribute_name, attribute_type, default_value
            )
            dict_dynamic_model_creation.update(dict_sme)
    model_type = create_model(
        class_name,
        **dict_dynamic_model_creation,
        __base__=aas_model.SubmodelElementCollection,
    )
    return model_type


def convert_submodel_list_to_pydantic_model(
    sm_element: model.SubmodelElementList, immutable: bool = False
) -> type[
    typing.Union[
        typing.List[aas_model.SubmodelElement], typing.Set[aas_model.SubmodelElement]
    ]
]:
    """
    Converts a SubmodelElementList to a Pydantic model.

    Args:
        sm_element (model.SubmodelElementList): SubmodelElementList to convert.

    Returns:
        typing.List[aas_model.SubmodelElement]: List of Pydantic models of the submodel elements.
    """
    if sm_element.value_type_list_element is not None:
        value_type = convert_xsdtype_to_primitive_type(
            sm_element.value_type_list_element
        )
    else:
        sm_element_value = sm_element.value[0]
        if isinstance(sm_element_value, model.SubmodelElementCollection):
            new_sm_element_value = unpatch_id_short_from_temp_attribute(
                sm_element_value
            )
            value_type = convert_submodel_collection_to_pydantic_model(
                new_sm_element_value
            )
            repatch_id_short_to_temp_attribute(sm_element_value, new_sm_element_value)

        elif isinstance(sm_element_value, model.SubmodelElementList):
            value_type = convert_submodel_list_to_pydantic_model(sm_element_value)
        elif isinstance(sm_element_value, model.ReferenceElement):
            value_type = convert_reference_element_to_pydantic_model(sm_element_value)
        elif isinstance(sm_element_value, model.RelationshipElement):
            value_type = convert_relationship_element_to_pydantic_model(
                sm_element_value
            )
        elif isinstance(sm_element_value, model.Property):
            value_type = convert_property_to_pydantic_model(sm_element_value)
        else:
            raise NotImplementedError("Type not implemented:", type(sm_element_value))
    if immutable:
        return typing.Tuple[value_type, ...]
    if not sm_element.order_relevant:
        return typing.Set[value_type]
    return typing.List[value_type]


def convert_reference_element_to_pydantic_model(
    sm_element: model.ReferenceElement,
) -> type[aas_model.Reference]:
    """
    Converts a ReferenceElement to a Pydantic model.

    Args:
        sm_element (model.ReferenceElement): ReferenceElement to convert.

    Returns:
        str: Value of the ReferenceElement.
    """
    return aas_model.Reference


def convert_relationship_element_to_pydantic_model(
    sm_element: model.RelationshipElement,
) -> type[typing.Tuple[aas_model.Reference, aas_model.Reference]]:
    """
    Converts a RelationshipElement to a Pydantic model.
    """
    return typing.Tuple[aas_model.Reference, aas_model.Reference]


def convert_property_to_pydantic_model(
    sm_element: model.Property,
) -> aas_model.PrimitiveSubmodelElement:
    """
    Converts a Property to a Pydantic model.

    Args:
        sm_element (model.Property): Property to convert.

    Returns:
        aas_model.PrimitiveSubmodelElement: Value of the Property.
    """
    return convert_xsdtype_to_primitive_type(sm_element.value_type)


def convert_multi_language_property_to_pydantic_model(
    sm_element: model.MultiLanguageProperty,
) -> aas_model.PrimitiveSubmodelElement:
    """
    Converts a MultiLanguageProperty to a Pydantic model.

    Args:
        sm_element (model.MultiLanguageProperty): MultiLanguageProperty to convert.

    Returns:
        aas_model.PrimitiveSubmodelElement: Value of the MultiLanguageProperty.
    """
    return str


def convert_file_to_pydantic_model(sm_element: model.File) -> type[aas_model.File]:
    """
    Convert a File to a pydantic model

    Args:
        sm_element (model.File): Basyx File to convert.

    Returns:
        aas_model.File: Pydantic model of the file
    """
    return aas_model.File


def convert_blob_to_pydantic_model(sm_element: model.Blob) -> type[aas_model.Blob]:
    """
    Convert a Blob to a pydantic model

    Args:
        sm_element (model.Blob): Basyx Blob to convert.

    Returns:
        aas_model.Blob: Pydantic model of the Blob
    """
    return aas_model.Blob



from __future__ import annotations

import typing
from pydantic import BaseModel, TypeAdapter

from aas_pydantic import aas_model, convert_aas_template, convert_util
from basyx.aas import model


from aas_pydantic.convert_util import (
    is_attribute_from_basyx_model_immutable,
    get_semantic_id_value_of_model,
    repatch_id_short_to_temp_attribute,
    unpatch_id_short_from_temp_attribute,
)


def get_types_name_dict(
    types: typing.List[typing.Type[aas_model.AAS | aas_model.Submodel]],
) -> typing.Dict[str, type]:
    """
    Returns a dictionary with the type names as keys and the types as values.

    Args:
        types (typing.List[type]): List of types to create the dictionary from.

    Returns:
        typing.Dict[str, type]: Dictionary with the type names as keys and the types as values.
    """
    types_name_dict = {t.__name__.split(".")[-1]: t for t in types}
    for top_level_type in types:
        if not issubclass(top_level_type, aas_model.AAS):
            continue
        for attribute_name, attribute_type in top_level_type.model_fields.items():
            if typing.get_origin(attribute_type.annotation) is typing.Union:
                for contained_type in typing.get_args(attribute_type.annotation):
                    if typing.get_origin(contained_type) is typing.Annotated:
                        contained_type = typing.get_args(contained_type)[0]
                    if not issubclass(contained_type, aas_model.Submodel):
                        continue
                    types_name_dict[contained_type.__name__.split(".")[-1]] = (
                        contained_type
                    )
                continue
            if not issubclass(attribute_type.annotation, aas_model.Submodel):
                continue
            types_name_dict[attribute_type.annotation.__name__.split(".")[-1]] = (
                attribute_type.annotation
            )
    return types_name_dict


def convert_object_store_to_pydantic_models(
    obj_store: model.DictObjectStore, types: typing.List[type]
) -> typing.List[aas_model.AAS]:
    """
    Converts an object store with AAS and submodels to pydantic models, representing the original data structure.

    Args:
        obj_store (model.DictObjectStore): Object store with AAS and submodels
        types (typing.List[type]): List of types to create the pydantic models from. Can be only top level types.

    Returns:
        typing.List[aas_model.AAS]: List of pydantic models
    """
    type_name_dict = get_types_name_dict(types)

    pydantic_submodels: typing.List[aas_model.Submodel] = []
    for identifiable in obj_store:
        if not isinstance(identifiable, model.Submodel):
            continue
        class_name = convert_util.get_class_name_from_basyx_model(identifiable)
        if not class_name in type_name_dict:
            pass
        pydantic_submodel = convert_submodel_to_model_instance(
            identifiable, type_name_dict[class_name]
        )
        pydantic_submodels.append(pydantic_submodel)

    pydantic_aas_list: typing.List[aas_model.AAS] = []
    for identifiable in obj_store:
        if not isinstance(identifiable, model.AssetAdministrationShell):
            continue
        class_name = convert_util.get_class_name_from_basyx_model(identifiable)
        if not class_name in type_name_dict:
            pass
        pydantic_aas = convert_aas_to_pydantic_model_instance(
            identifiable, pydantic_submodels, type_name_dict[class_name]
        )
        pydantic_aas_list.append(pydantic_aas)

    return pydantic_aas_list


def convert_aas_to_pydantic_model_instance(
    aas: model.AssetAdministrationShell,
    pydantic_submodels: typing.List[aas_model.Submodel],
    model_type: type = aas_model.AAS,
) -> aas_model.AAS:
    """
    Converts an AAS to a Pydantic model.

    Args:
        aas (model.AssetAdministrationShell): AAS to convert

    Returns:
        aas_model.AAS: Pydantic model of the asset administration shell
    """
    dict_model_instantiation = get_initial_dict_for_model_instantiation(aas)
    aas_submodel_ids = [sm.get_identifier() for sm in aas.submodel]

    for sm in pydantic_submodels:
        if not sm.id in aas_submodel_ids:
            continue
        attribute_names_of_submodel = (
            convert_util.get_attribute_names_from_basyx_template(aas, sm.id)
        )
        if len(attribute_names_of_submodel) > 1:
            raise ValueError(
                "Multiple attribute names found for submodel:",
                attribute_names_of_submodel,
            )
        attribute_name_of_submodel = attribute_names_of_submodel[0]
        dict_model_instantiation.update({attribute_name_of_submodel: sm.model_dump()})
    return TypeAdapter(model_type).validate_python(dict_model_instantiation)


def get_submodel_element_value(
    sm_element: model.SubmodelElement, attribute_type: type = None
) -> aas_model.SubmodelElement:
    """
    Returns the value of a SubmodelElement.

    Args:
        sm_element (model.SubmodelElement): SubmodelElement to get the value from.

    Returns:
        aas_model.SubmodelElement: Value of the SubmodelElement.
    """
    if isinstance(sm_element, model.SubmodelElementCollection):
        return convert_submodel_collection_to_pydantic_model(
            sm_element, model_type=attribute_type
        )
    elif isinstance(sm_element, model.SubmodelElementList):
        return convert_submodel_list_to_pydantic_model(
            sm_element, model_type=attribute_type
        )
    elif isinstance(sm_element, model.ReferenceElement):
        return convert_reference_element_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.Property):
        return convert_property_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.File):
        return convert_file_to_pydantic_model(sm_element)
    elif isinstance(sm_element, model.Blob):
        return convert_blob_to_pydantic_model(sm_element)
    else:
        raise NotImplementedError("Type not implemented:", type(sm_element))


def get_model_instantiation_dict_from_submodel_element(
    attribute_name: str, attribute_value: typing.Any
) -> typing.Dict[str, typing.Any]:
    """
    Converts a SubmodelElement to a dict.

    Args:
        attribute_name (str): Name of the attribute to create in the dictionary.
        sm_element (model.SubmodelElement): SubmodelElement to convert.

    Returns:
        dict: Dictionary that can be used to instantiate a Pydantic model.
    """
    if isinstance(attribute_value, BaseModel):
        attribute_value = attribute_value.model_dump()
    elif isinstance(attribute_value, (list, set, tuple)) and any(
        isinstance(element, BaseModel) for element in attribute_value
    ):
        attribute_value = [element.model_dump() for element in attribute_value]
    return {attribute_name: attribute_value}


def get_initial_dict_for_model_instantiation(
    basyx_model: (
        model.Submodel
        | model.AssetAdministrationShell
        | model.SubmodelElementCollection
    ),
) -> typing.Dict[str, typing.Any]:
    """
    Returns a dictionary that can be used to instantiate a Pydantic model based on a provided basyx submodel.

    Args:
        basyx_model (model.Submodel | model.AssetAdministrationShell | model.SubmodelElementCollection): Basyx model to create the dictionary from.

    Returns:
        typing.Dict[str, typing.Any]: Dictionary that can be used to instantiate a Pydantic model.
    """
    model_instantiation_dict = {
        "id_short": basyx_model.id_short,
        "description": convert_util.get_str_description(basyx_model.description),
    }
    if isinstance(basyx_model, model.Identifiable):
        model_instantiation_dict["id"] = str(basyx_model.id)
    if isinstance(basyx_model, model.HasSemantics):
        model_instantiation_dict["semantic_id"] = get_semantic_id_value_of_model(
            basyx_model
        )
    return model_instantiation_dict


def convert_submodel_to_model_instance(
    sm: model.Submodel, model_type: type[aas_model.Submodel] = None
) -> aas_model.Submodel:
    """
    Converts a Submodel to a Pydantic model.

    Args:
        sm (model.Submodel): Submodel to convert.
        model_type (type[aas_model.Submodel]): Pydantic model type to convert the submodel to.

    Returns:
        aas_model.Submodel: Pydantic model of the submodel.
    """
    if model_type is None:
        model_type = convert_aas_template.convert_submodel_template_to_pydatic_type(sm)
    dict_model_instantiation = get_initial_dict_for_model_instantiation(sm)

    for sm_element in sm.submodel_element:
        attribute_names = convert_util.get_attribute_names_from_basyx_template(
            sm, sm_element.id_short
        )
        if len(attribute_names) > 1:
            raise ValueError(
                "Multiple attribute names found for submodel element:", attribute_names
            )
        attribute_name = attribute_names[0]
        attribute_type = get_type_of_attribute(model_type, attribute_name)
        attribute_value = get_submodel_element_value(sm_element, attribute_type)
        sme_model_instantiation_dict = (
            get_model_instantiation_dict_from_submodel_element(
                attribute_name, attribute_value
            )
        )
        dict_model_instantiation.update(sme_model_instantiation_dict)
    return TypeAdapter(model_type).validate_python(dict_model_instantiation)


def get_type_of_attribute(
    model_type: typing.Union[type[BaseModel], typing.Union[type[BaseModel]]],
    attribute_name: str,
) -> typing.Union[type[BaseModel]]:
    """
    Returns the type of an attribute of a Pydantic model.

    Args:
        model_type (typing.Union[type[BaseModel], typing.Union]): Pydantic model type to get the attribute type from.
        attribute_name (str): Name of the attribute to get the type from.

    Returns:
        typing.Union[type[BaseModel], typing.Union]: Type of the attribute.
    """
    if typing.get_origin(model_type) is typing.Union:
        for contained_type in typing.get_args(model_type):
            if (
                hasattr(contained_type, "model_fields")
                and attribute_name in contained_type.model_fields
            ):
                return contained_type.model_fields[attribute_name].annotation
    if (
        hasattr(model_type, "model_fields")
        and attribute_name in model_type.model_fields
    ):
        return model_type.model_fields[attribute_name].annotation
    raise ValueError(
        f"Attribute {attribute_name} not found in model fields with attributes {list(model_type.model_fields.keys())}."
    )


def convert_submodel_collection_to_pydantic_model(
    sm_element: model.SubmodelElementCollection,
    model_type: type[aas_model.SubmodelElementCollection],
) -> aas_model.SubmodelElementCollection:
    """
    Converts a SubmodelElementCollection to a Pydantic model.

    Args:
        sm_element (model.SubmodelElementCollection): SubmodelElementCollection to convert.

    Returns:
        aas_model.SubmodelElementCollection: Pydantic model of the submodel element collection.
    """
    dict_model_instantiation = get_initial_dict_for_model_instantiation(sm_element)

    for sub_sm_element in sm_element.value:
        attribute_names = convert_util.get_attribute_names_from_basyx_template(
            sm_element, sub_sm_element.id_short
        )
        if len(attribute_names) > 1:
            raise ValueError(
                "Multiple attribute names found for submodel element:", attribute_names
            )
        attribute_name = attribute_names[0]
        attribute_type = get_type_of_attribute(model_type, attribute_name)
        attribute_value = get_submodel_element_value(sub_sm_element, attribute_type)
        dict_sme_instantiation = get_model_instantiation_dict_from_submodel_element(
            attribute_name, attribute_value
        )
        dict_model_instantiation.update(dict_sme_instantiation)
    return TypeAdapter(model_type).validate_python(dict_model_instantiation)


def convert_submodel_list_to_pydantic_model(
    sm_element: model.SubmodelElementList, model_type: type[typing.Any]
) -> typing.Union[
    typing.List[aas_model.SubmodelElement],
    typing.Set[aas_model.SubmodelElement],
    typing.Tuple[aas_model.SubmodelElement],
]:
    """
    Converts a SubmodelElementList to a Pydantic model.

    Args:
        sm_element (model.SubmodelElementList): SubmodelElementList to convert.

    Returns:
        typing.List[aas_model.SubmodelElement]: List of Pydantic models of the submodel elements.
    """
    sme_pydantic_models = []
    for sme in sm_element.value:
        if isinstance(sme, model.SubmodelElementCollection):
            new_sme = unpatch_id_short_from_temp_attribute(sme)
            sme_pydantic_models.append(
                convert_submodel_collection_to_pydantic_model(
                    new_sme, model_type=typing.get_args(model_type)[0]
                )
            )
            repatch_id_short_to_temp_attribute(sme, new_sme)
        elif isinstance(sme, model.SubmodelElementList):
            sme_pydantic_models.append(
                convert_submodel_list_to_pydantic_model(
                    sme, model_type=typing.get_args(model_type)[0]
                )
            )
        elif isinstance(sme, model.ReferenceElement):
            sme_pydantic_models.append(convert_reference_element_to_pydantic_model(sme))
        elif isinstance(sme, model.Property):
            sme_pydantic_models.append(sme.value)
        else:
            raise NotImplementedError("Type not implemented:", type(sme))
    if not sm_element.order_relevant:
        return set(sme_pydantic_models)
    if model_type is tuple:
        return tuple(sme_pydantic_models)
    return sme_pydantic_models


def convert_reference_element_to_pydantic_model(
    sm_element: model.ReferenceElement,
) -> str:
    """
    Converts a ReferenceElement to a Pydantic model.

    Args:
        sm_element (model.ReferenceElement): ReferenceElement to convert.

    Returns:
        str: Value of the ReferenceElement.
    """
    return sm_element.value.key[0].value


def convert_property_to_pydantic_model(
    sm_element: model.Property,
) -> aas_model.PrimitiveSubmodelElement:
    """
    Converts a Property to a Pydantic model.

    Args:
        sm_element (model.Property): Property to convert.

    Returns:
        aas_model.PrimitiveSubmodelElement: Value of the Property.
    """
    return sm_element.value


def convert_file_to_pydantic_model(sm_element: model.File) -> aas_model.File:
    """
    Convert a File to a pydantic model

    Args:
        sm_element (model.File): Basyx File to convert.

    Returns:
        aas_model.File: Pydantic model of the file
    """
    return aas_model.File(
        id_short=sm_element.id_short,
        description=sm_element.description if sm_element.semantic_id else "",
        semantic_id=sm_element.semantic_id if sm_element.semantic_id else "",
        media_type=sm_element.content_type,
        path=sm_element.value,
    )


def convert_blob_to_pydantic_model(sm_element: model.Blob) -> aas_model.Blob:
    """
    Convert a Blob to a pydantic model

    Args:
        sm_element (model.Blob): Basyx Blob to convert.

    Returns:
        aas_model.Blob: Pydantic model of the Blob
    """
    return aas_model.Blob(
        id_short=sm_element.id_short,
        description=sm_element.description if sm_element.semantic_id else "",
        semantic_id=sm_element.semantic_id if sm_element.semantic_id else "",
        media_type=sm_element.content_type,
        content=sm_element.value,
    )
